from pwn import *
from sage.all import *

x = var('x')

class HSM:
	def __init__(self):
		modulus = x**0 + x**1 + x**2 + x**4 + x**5 + x**7 + x**10 + x**11 + x**12 + x**13 + x**18 + x**20 + x**21 + x**22 + x**23 + x**24 + x**25 + x**26 + x**30 + x**33 + x**64
		self.k = GF(2**64,name='x',modulus=modulus)
		
	def find_keys(self,plaintextA,plaintextB,signA,signB):
		
		msgA_gf = self.k._cache.fetch_int(Integer(plaintextA))
		msgB_gf = self.k._cache.fetch_int(Integer(plaintextB))
		sigA_gf = self.k._cache.fetch_int(Integer(signA))
		sigB_gf = self.k._cache.fetch_int(Integer(signB))
		
		K0s_gf = ((sigA_gf - sigB_gf) / (msgA_gf - msgB_gf)).nth_root(3,all=True)
		
		solutions = []
		for K0_gf in K0s_gf:
			K1_gf = (sigB_gf / K0_gf) - (msgB_gf * (K0_gf * K0_gf))
			
			K0 = int(str(''.join(map(str,K0_gf.polynomial())))[::-1],2)
			K1 = int(str(''.join(map(str,K1_gf.polynomial())))[::-1],2)
			solutions.append( (K0, K1))
		
		return solutions
		
	def set_keys(self,K0,K1):
		self.K0_gf = self.k._cache.fetch_int(Integer(K0))
		self.K1_gf = self.k._cache.fetch_int(Integer(K1))
		
	def sign_u64(self,v1,v2):
		v1_gf = self.k._cache.fetch_int(Integer(v1))
		v2_gf = self.k._cache.fetch_int(Integer(v2))
		
		X_gf = self.K0_gf * v1_gf
		X_gf = v2_gf + X_gf
		X_gf = self.K0_gf * X_gf
		X_gf = X_gf + self.K1_gf
		X_gf = self.K0_gf * X_gf
		
		return int(str(''.join(map(str,X_gf.polynomial())))[::-1],2)
		
	def sign_data(self,plaintext):
		sign = 0
		padded = (int(len(plaintext)/8) + int(len(plaintext)%8!=0) )*8
		plaintext = plaintext + ( b"\x00" * (padded - len(plaintext)) ) 
		for i in range(0,len(plaintext),8):
			value = u64(plaintext[i:i+8])
			sign = self.sign_u64(value,sign)
		return sign
	
	def sign_pointer(self,value):
		signature = self.sign_u64(value & 0x0000FFFFFFFFFFFF,0) & 0xFFFF000000000000
		return value | signature

	def auth_pointer(self,value):
		pointer = value & 0x0000FFFFFFFFFFFF
		if self.sign_u64(pointer,0) & 0xFFFF000000000000 == value & 0xFFFF000000000000:
			return pointer
		else:
			raise Exception("Invalid signature")