import ctypes
import argparse
from pwn import *

class HEADER(ctypes.LittleEndianStructure):
	_fields_ = [
		("Signature",ctypes.c_uint8 * 8),
		("CLSID",ctypes.c_uint8 * 16),
		("MinorVersion",ctypes.c_uint16),
		("MajorVersion",ctypes.c_uint16),
		("ByteOrder",ctypes.c_uint16),
		("SectorShift",ctypes.c_uint16),
		("MiniSectorShift",ctypes.c_uint16),
		("Reserved",ctypes.c_uint8 * 6),
		("NumberOfDirectorySectors",ctypes.c_uint32),
		("NumberOfFATSectors",ctypes.c_uint32),
		("FirstDirectorySectorLocation",ctypes.c_uint32),
		("TransactionSignatureNumber",ctypes.c_uint32),
		("MiniStreamCutOffSize",ctypes.c_uint32),
		("FirstMiniFATSectorLocation",ctypes.c_uint32),
		("NumberOfMiniFATSectors",ctypes.c_uint32),
		("FirstDIFATSectorLocation",ctypes.c_uint32),
		("NumberOfDIFATSectors",ctypes.c_uint32),
		("DIFAT",ctypes.c_uint32 * 109)
	]
	
def extract(path,output_path,offset):
	with open(path,"rb") as f:
		# Extract FAT
		header = HEADER.from_buffer_copy(f.read(ctypes.sizeof(HEADER)))
		fat = []
		for i in range(0,109):
			if header.DIFAT[i] != 0xFFFFFFFF:
				sector_offset = (header.DIFAT[i] + 1) * 512
				f.seek(sector_offset)
				data = f.read(512)
				for k in range(0,len(data),4):
					fat.append(u32(data[k:k+4]))
		# Extract file
		with open(output_path,"wb") as fout:
			linked_list = []
			fat_index = int(offset/512)-1
			linked_list.append(fat_index)
			f.seek( (fat_index + 1)*512 )
			fout.write(f.read(512))
			while fat_index < 0xFFFFFFFC and fat_index < len(fat):
				fat_index = fat[fat_index]
				if fat_index in linked_list:
					break
				linked_list.append(fat_index)
				f.seek( (fat_index + 1)*512 )
				fout.write(f.read(512))
			
	
if __name__=='__main__':
	parser = argparse.ArgumentParser(description="extract hidden file from FAT in .doc")
	parser.add_argument('input',type=str,help='doc file')
	parser.add_argument('output',type=str,help='output file')
	parser.add_argument('offset',type=str,help='hidden file begin')
	args = parser.parse_args()
	
	args.offset = int(args.offset,16)
	
	extract(args.input,args.output,args.offset)